# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
4300 - Module for testing other cursor methods and attributes.
"""

import decimal
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def test_4300(self):
        "4300 - test preparing a statement and executing it multiple times"
        cursor = self.conn.cursor()
        self.assertEqual(cursor.statement, None)
        statement = "begin :value := :value + 5; end;"
        cursor.prepare(statement)
        var = cursor.var(oracledb.NUMBER)
        self.assertEqual(cursor.statement, statement)
        var.setvalue(0, 2)
        cursor.execute(None, value=var)
        self.assertEqual(var.getvalue(), 7)
        cursor.execute(None, value=var)
        self.assertEqual(var.getvalue(), 12)
        cursor.execute("begin :value2 := 3; end;", value2=var)
        self.assertEqual(var.getvalue(), 3)

    def test_4301(self):
        "4301 - confirm an exception is raised after closing a cursor"
        self.cursor.close()
        with self.assertRaisesFullCode("DPY-1006"):
            self.cursor.execute("select 1 from dual")

    def test_4302(self):
        "4302 - test iterators"
        self.cursor.execute(
            """
            select IntCol
            from TestNumbers
            where IntCol between 1 and 3
            order by IntCol
            """
        )
        rows = [v for v, in self.cursor]
        self.assertEqual(rows, [1, 2, 3])

    def test_4303(self):
        "4303 - test iterators (with intermediate execute)"
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.execute(
            """
            select IntCol
            from TestNumbers
            where IntCol between 1 and 3
            order by IntCol
            """
        )
        test_iter = iter(self.cursor)
        (value,) = next(test_iter)
        self.cursor.execute("insert into TestTempTable (IntCol) values (1)")
        with self.assertRaisesFullCode("DPY-1003"):
            next(test_iter)

    def test_4304(self):
        "4304 - test that bindnames() works correctly."
        cursor = self.conn.cursor()
        with self.assertRaisesFullCode("DPY-2002"):
            cursor.bindnames()
        cursor.prepare("begin null; end;")
        self.assertEqual(cursor.bindnames(), [])
        cursor.prepare("begin :retval := :inval + 5; end;")
        self.assertEqual(cursor.bindnames(), ["RETVAL", "INVAL"])
        cursor.prepare("begin :retval := :a * :a + :b * :b; end;")
        self.assertEqual(cursor.bindnames(), ["RETVAL", "A", "B"])
        cursor.prepare(
            """
            begin
                :a := :b + :c + :d + :e + :f + :g + :h + :i + :j + :k + :l;
            end;
            """
        )
        names = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L"]
        self.assertEqual(cursor.bindnames(), names)
        cursor.prepare("select :a * :a + :b * :b from dual")
        self.assertEqual(cursor.bindnames(), ["A", "B"])
        cursor.prepare("select :value1 + :VaLue_2 from dual")
        self.assertEqual(cursor.bindnames(), ["VALUE1", "VALUE_2"])
        cursor.prepare("select :élevé, :fenêtre from dual")
        self.assertEqual(cursor.bindnames(), ["ÉLEVÉ", "FENÊTRE"])

    def test_4305(self):
        "4305 - test cursor.setinputsizes() with invalid parameters"
        val = decimal.Decimal(5)
        with self.assertRaisesFullCode("DPY-2005"):
            self.cursor.setinputsizes(val, x=val)
        with self.assertRaisesFullCode("DPY-2007"):
            self.cursor.setinputsizes(val)

    def test_4306(self):
        "4306 - test setting input sizes without any parameters"
        self.cursor.setinputsizes()
        self.cursor.execute("select :val from dual", val="Test Value")
        self.assertEqual(self.cursor.fetchall(), [("Test Value",)])

    def test_4307(self):
        "4307 - test setting input sizes with an empty dictionary"
        empty_dict = {}
        self.cursor.prepare("select 236 from dual")
        self.cursor.setinputsizes(**empty_dict)
        self.cursor.execute(None, empty_dict)
        self.assertEqual(self.cursor.fetchall(), [(236,)])

    def test_4308(self):
        "4308 - test setting input sizes with an empty list"
        empty_list = []
        self.cursor.prepare("select 239 from dual")
        self.cursor.setinputsizes(*empty_list)
        self.cursor.execute(None, empty_list)
        self.assertEqual(self.cursor.fetchall(), [(239,)])

    def test_4309(self):
        "4309 - test setting input sizes with positional args"
        var = self.cursor.var(oracledb.STRING, 100)
        self.cursor.setinputsizes(None, 5, None, 10, None, oracledb.NUMBER)
        self.cursor.execute(
            """
            begin
              :1 := :2 || to_char(:3) || :4 || to_char(:5) || to_char(:6);
            end;
            """,
            [var, "test_", 5, "_second_", 3, 7],
        )
        self.assertEqual(var.getvalue(), "test_5_second_37")

    def test_4310(self):
        "4310 - test Cursor repr()"
        expected_value = f"<oracledb.Cursor on {self.conn}>"
        self.assertEqual(str(self.cursor), expected_value)
        self.assertEqual(repr(self.cursor), expected_value)

    def test_4311(self):
        "4311 - test parsing query statements"
        sql = "select LongIntCol from TestNumbers where IntCol = :val"
        self.cursor.parse(sql)
        self.assertEqual(self.cursor.statement, sql)
        self.assertEqual(
            self.cursor.description,
            [("LONGINTCOL", oracledb.DB_TYPE_NUMBER, 17, None, 16, 0, 0)],
        )

    def test_4312(self):
        "4312 - test cursor.setoutputsize() does not fail (but does nothing)"
        self.cursor.setoutputsize(100, 2)

    def test_4313(self):
        "4313 - test cursor.var() with invalid parameters"
        with self.assertRaisesFullCode("DPY-2007"):
            self.cursor.var(5)

    def test_4314(self):
        "4314 - test cursor.arrayvar() with invalid parameters"
        with self.assertRaisesFullCode("DPY-2007"):
            self.cursor.arrayvar(5, 1)

    def test_4315(self):
        "4315 - test binding boolean data without the use of PL/SQL"
        self.cursor.execute("truncate table TestTempTable")
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        self.cursor.execute(sql, (False, "Value should be 0"))
        self.cursor.execute(sql, (True, "Value should be 1"))
        self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable order by IntCol"
        )
        expected_value = [(0, "Value should be 0"), (1, "Value should be 1")]
        self.assertEqual(self.cursor.fetchall(), expected_value)

    def test_4316(self):
        "4316 - test using a cursor as a context manager"
        with self.cursor as cursor:
            cursor.execute("truncate table TestTempTable")
            cursor.execute("select count(*) from TestTempTable")
            (count,) = cursor.fetchone()
            self.assertEqual(count, 0)
        with self.assertRaisesFullCode("DPY-1006"):
            self.cursor.close()

    def test_4317(self):
        "4317 - test that rowcount attribute is reset to zero on query execute"
        for num in [0, 1, 1, 0]:
            self.cursor.execute("select * from dual where 1 = :s", [num])
            self.cursor.fetchone()
            self.assertEqual(self.cursor.rowcount, num)

    def test_4318(self):
        "4318 - test that the typename attribute can be passed a value of None"
        value_to_set = 5
        var = self.cursor.var(int, typename=None)
        var.setvalue(0, value_to_set)
        self.assertEqual(var.getvalue(), value_to_set)

    def test_4319(self):
        "4319 - test that an object type can be used as type in cursor.var()"
        obj_type = self.conn.gettype("UDT_OBJECT")
        var = self.cursor.var(obj_type)
        self.cursor.callproc(
            "pkg_TestBindObject.BindObjectOut", (28, "Bind obj out", var)
        )
        obj = var.getvalue()
        result = self.cursor.callfunc(
            "pkg_TestBindObject.GetStringRep", str, (obj,)
        )
        exp = "udt_Object(28, 'Bind obj out', null, null, null, null, null)"
        self.assertEqual(result, exp)

    def test_4320(self):
        "4320 - test that fetching an XMLType returns a string"
        int_val = 5
        label = "IntCol"
        expected_result = f"<{label}>{int_val}</{label}>"
        self.cursor.execute(
            f"""
            select XMLElement("{label}", IntCol)
            from TestStrings
            where IntCol = :int_val
            """,
            int_val=int_val,
        )
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, expected_result)

    def test_4321(self):
        "4321 - test last rowid"

        # no statement executed: no rowid
        self.assertIsNone(self.cursor.lastrowid)

        # DDL statement executed: no rowid
        self.cursor.execute("truncate table TestTempTable")
        self.assertIsNone(self.cursor.lastrowid)

        # statement prepared: no rowid
        self.cursor.prepare("insert into TestTempTable (IntCol) values (:1)")
        self.assertIsNone(self.cursor.lastrowid)

        # multiple rows inserted: rowid of last row inserted
        rows = [(n,) for n in range(225)]
        self.cursor.executemany(None, rows)
        rowid = self.cursor.lastrowid
        self.cursor.execute(
            """
            select rowid
            from TestTempTable
            where IntCol = :1
            """,
            rows[-1],
        )
        self.assertEqual(self.cursor.fetchone()[0], rowid)

        # statement executed but no rows updated: no rowid
        self.cursor.execute("delete from TestTempTable where 1 = 0")
        self.assertIsNone(self.cursor.lastrowid)

        # stetement executed with one row updated: rowid of updated row
        self.cursor.execute(
            """
            update TestTempTable set StringCol1 = 'Modified'
            where IntCol = :1
            """,
            rows[-2],
        )
        rowid = self.cursor.lastrowid
        self.cursor.execute(
            "select rowid from TestTempTable where IntCol = :1",
            rows[-2],
        )
        self.assertEqual(self.cursor.fetchone()[0], rowid)

        # statement executed with many rows updated: rowid of last updated row
        self.cursor.execute(
            """
            update TestTempTable set
                StringCol1 = 'Row ' || to_char(IntCol)
            where IntCol = :1
            """,
            rows[-3],
        )
        rowid = self.cursor.lastrowid
        self.cursor.execute(
            "select StringCol1 from TestTempTable where rowid = :1",
            [rowid],
        )
        self.assertEqual(self.cursor.fetchone()[0], "Row %s" % rows[-3])

    def test_4322(self):
        "4322 - test prefetch rows"
        self.setup_round_trip_checker()

        # perform simple query and verify only one round trip is needed
        with self.conn.cursor() as cursor:
            cursor.execute("select sysdate from dual").fetchall()
            self.assertRoundTrips(1)

        # set prefetchrows to 1 and verify that two round trips are now needed
        with self.conn.cursor() as cursor:
            cursor.prefetchrows = 1
            self.assertEqual(cursor.prefetchrows, 1)
            cursor.execute("select sysdate from dual").fetchall()
            self.assertRoundTrips(2)

        # simple DDL only requires a single round trip
        with self.conn.cursor() as cursor:
            cursor.execute("truncate table TestTempTable")
            self.assertRoundTrips(1)

        # array execution only requires a single round trip
        num_rows = 590
        with self.conn.cursor() as cursor:
            data = [(n + 1,) for n in range(num_rows)]
            cursor.executemany(
                "insert into TestTempTable (IntCol) values (:1)",
                data,
            )
            self.assertRoundTrips(1)

        # setting prefetch and array size to 1 requires a round-trip for each
        # row
        with self.conn.cursor() as cursor:
            cursor.arraysize = 1
            cursor.prefetchrows = 1
            self.assertEqual(cursor.prefetchrows, 1)
            cursor.execute("select IntCol from TestTempTable").fetchall()
            self.assertRoundTrips(num_rows + 1)

        # setting prefetch and array size to 300 requires 2 round-trips
        with self.conn.cursor() as cursor:
            cursor.arraysize = 300
            cursor.prefetchrows = 300
            self.assertEqual(cursor.prefetchrows, 300)
            cursor.execute("select IntCol from TestTempTable").fetchall()
            self.assertRoundTrips(2)

    def test_4323(self):
        "4323 - test prefetch rows using existing cursor"
        self.setup_round_trip_checker()

        # Set prefetch rows on an existing cursor
        num_rows = 590
        with self.conn.cursor() as cursor:
            cursor.execute("truncate table TestTempTable")
            self.assertRoundTrips(1)
            data = [(n + 1,) for n in range(num_rows)]
            cursor.executemany(
                "insert into TestTempTable (IntCol) values (:1)",
                data,
            )
            self.assertRoundTrips(1)
            cursor.prefetchrows = 30
            cursor.arraysize = 100
            cursor.execute("select IntCol from TestTempTable").fetchall()
            self.assertRoundTrips(7)

    def test_4324(self):
        "4324 - test parsing plsql statements"
        sql = "begin :value := 5; end;"
        self.cursor.parse(sql)
        self.assertEqual(self.cursor.statement, sql)
        self.assertIsNone(self.cursor.description)

    def test_4325(self):
        "4325 - test parsing ddl statements"
        sql = "truncate table TestTempTable"
        self.cursor.parse(sql)
        self.assertEqual(self.cursor.statement, sql)
        self.assertIsNone(self.cursor.description)

    def test_4326(self):
        "4326 - test parsing dml statements"
        sql = "insert into TestTempTable (IntCol) values (1)"
        self.cursor.parse(sql)
        self.assertEqual(self.cursor.statement, sql)
        self.assertIsNone(self.cursor.description)

    def test_4327(self):
        "4327 - test to verify encodingErrors is deprecated"
        errors = "strict"
        with self.assertRaisesFullCode("DPY-2014"):
            self.cursor.var(
                oracledb.NUMBER, encoding_errors=errors, encodingErrors=errors
            )

    def test_4328(self):
        "4328 - test arrays of arrays not supported"
        simple_var = self.cursor.arrayvar(oracledb.NUMBER, 3)
        with self.assertRaisesFullCode("DPY-3005"):
            simple_var.setvalue(1, [1, 2, 3])

    def test_4329(self):
        "4329 - test cursor.setinputsizes() with invalid list parameters"
        with self.assertRaisesFullCode("DPY-2011"):
            self.cursor.setinputsizes([int, 2, 10])

    def test_4330(self):
        "4330 - test unsupported python type on cursor"
        with self.assertRaisesFullCode("DPY-3003"):
            self.cursor.var(list)

    def test_4331(self):
        "4331 - test binding by name with leading colon"
        params = {":arg1": 5}
        self.cursor.execute("select :arg1 from dual", params)
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, params[":arg1"])

    def test_4332(self):
        "4332 - test binding mixed null and not null values in a PL/SQL block"
        out_vars = [self.cursor.var(str) for i in range(4)]
        self.cursor.execute(
            """
            begin
                :1 := null;
                :2 := 'Value 1';
                :3 := null;
                :4 := 'Value 2';
            end;
            """,
            out_vars,
        )
        values = [var.getvalue() for var in out_vars]
        self.assertEqual(values, [None, "Value 1", None, "Value 2"])

    def test_4333(self):
        "4333 - test excluding statement from statement cache"
        num_iters = 10
        sql = "select user from dual"
        self.setup_parse_count_checker()

        # with statement cache enabled, only one parse should take place
        for i in range(num_iters):
            with self.conn.cursor() as cursor:
                cursor.execute(sql)
        self.assertParseCount(1)

        # with statement cache disabled for the statement, parse count should
        # be the same as the number of iterations
        for i in range(num_iters):
            with self.conn.cursor() as cursor:
                cursor.prepare(sql, cache_statement=False)
                cursor.execute(None)
        self.assertParseCount(num_iters - 1)

    def test_4334(self):
        "4334 - test repeated DDL"
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.execute("insert into TestTempTable (IntCol) values (1)")
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.execute("insert into TestTempTable (IntCol) values (1)")

    def test_4335(self):
        "4335 - test executing SQL with non-ASCII characters"
        self.cursor.execute("select 'FÖÖ' from dual")
        (result,) = self.cursor.fetchone()
        self.assertIn(result, ("FÖÖ", "F¿¿"))

    def test_4336(self):
        "4336 - test case sensitivity of unquoted bind names"
        self.cursor.execute("select :test from dual", {"TEST": "a"})
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, "a")

    def test_4337(self):
        "4337 - test case sensitivity of quoted bind names"
        with self.assertRaisesFullCode("ORA-01036", "DPY-4008"):
            self.cursor.execute('select :"test" from dual', {'"TEST"': "a"})

    def test_4338(self):
        "4338 - test using a reserved keywords as a bind name"
        sql = "select :ROWID from dual"
        with self.assertRaisesFullCode("ORA-01745"):
            self.cursor.parse(sql)

    def test_4339(self):
        "4339 - test array size less than prefetch rows"
        for i in range(2):
            with self.conn.cursor() as cursor:
                cursor.arraysize = 1
                cursor.execute("select 1 from dual union select 2 from dual")
                self.assertEqual(cursor.fetchall(), [(1,), (2,)])

    def test_4340(self):
        "4340 - test re-executing a query with blob as bytes"

        def type_handler(cursor, metadata):
            if metadata.type_code is oracledb.DB_TYPE_BLOB:
                return cursor.var(bytes, arraysize=cursor.arraysize)

        self.conn.outputtypehandler = type_handler
        blob_data = b"An arbitrary set of blob data for test case 4348"
        self.cursor.execute("delete from TestBLOBs")
        self.cursor.execute(
            "insert into TestBLOBs (IntCol, BlobCol) values (1, :data)",
            [blob_data],
        )
        self.cursor.execute("select IntCol, BlobCol from TestBLOBs")
        self.assertEqual(self.cursor.fetchall(), [(1, blob_data)])

        self.cursor.execute("delete from TestBLOBs")
        self.cursor.execute(
            "insert into TestBLOBs (IntCol, BlobCol) values (1, :data)",
            [blob_data],
        )
        self.cursor.execute("select IntCol, BlobCol from TestBLOBs")
        self.assertEqual(self.cursor.fetchall(), [(1, blob_data)])

    def test_4341(self):
        "4341 - test re-executing a statement after raising an error"
        sql = "select * from TestFakeTable"
        with self.assertRaisesFullCode("ORA-00942"):
            self.cursor.execute(sql)
        with self.assertRaisesFullCode("ORA-00942"):
            self.cursor.execute(sql)

        sql = "insert into TestStrings (StringCol) values (NULL)"
        with self.assertRaisesFullCode("ORA-01400"):
            self.cursor.execute(sql)
        with self.assertRaisesFullCode("ORA-01400"):
            self.cursor.execute(sql)

    def test_4342(self):
        "4342 - test executing a statement that raises ORA-01007"
        with self.conn.cursor() as cursor:
            cursor.execute(
                """
                create or replace view ora_1007 as
                    select 1 as SampleNumber, 'String' as SampleString,
                        'Another String' as AnotherString
                    from dual
                """
            )
        with self.conn.cursor() as cursor:
            cursor.execute("select * from ora_1007")
            self.assertEqual(
                cursor.fetchone(), (1, "String", "Another String")
            )
        with self.conn.cursor() as cursor:
            cursor.execute(
                """
                create or replace view ora_1007 as
                    select 1 as SampleNumber,
                        'Another String' as AnotherString
                    from dual
                """
            )
        with self.conn.cursor() as cursor:
            cursor.execute("select * from ora_1007")
            self.assertEqual(cursor.fetchone(), (1, "Another String"))

    def test_4343(self):
        "4343 - test updating an empty row"
        int_var = self.cursor.var(int)
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.execute(
            """
            begin
                update TestTempTable set IntCol = :1
                where StringCol1 = :2
                returning IntCol into :3;
            end;
            """,
            [1, "test string 4352", int_var],
        )
        self.assertEqual(int_var.values, [None])

    def test_4344(self):
        "4344 - fetch duplicate data from query in statement cache"
        sql = """
                select 'A', 'B', 'C' from dual
                union all
                select 'A', 'B', 'C' from dual
                union all
                select 'A', 'B', 'C' from dual"""
        expected_data = [("A", "B", "C")] * 3
        with self.conn.cursor() as cursor:
            cursor.prefetchrows = 0
            cursor.execute(sql)
            self.assertEqual(cursor.fetchall(), expected_data)
        with self.conn.cursor() as cursor:
            cursor.prefetchrows = 0
            cursor.execute(sql)
            self.assertEqual(cursor.fetchall(), expected_data)

    def test_4345(self):
        "4345 - fetch duplicate data with outconverter"

        def out_converter(value):
            self.assertIs(type(value), str)
            return int(value)

        def type_handler(cursor, metadata):
            if metadata.name == "COL_3":
                return cursor.var(
                    str, arraysize=cursor.arraysize, outconverter=out_converter
                )

        self.cursor.outputtypehandler = type_handler
        self.cursor.execute(
            """
            select 'A' as col_1, 2 as col_2, 3 as col_3 from dual
                union all
            select 'A' as col_1, 2 as col_2, 3 as col_3 from dual
                union all
            select 'A' as col_1, 2 as col_2, 3 as col_3 from dual
            """
        )
        expected_data = [("A", 2, 3)] * 3
        self.assertEqual(self.cursor.fetchall(), expected_data)

    def test_4346(self):
        "4346 - test setinputsizes() with defaults specified"
        self.cursor.setinputsizes(None, str)
        self.assertIsNone(self.cursor.bindvars[0])
        self.assertIsInstance(self.cursor.bindvars[1], oracledb.Var)
        self.cursor.setinputsizes(a=None, b=str)
        self.assertIsNone(self.cursor.bindvars.get("a"))
        self.assertIsInstance(self.cursor.bindvars["b"], oracledb.Var)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_4347(self):
        "4547 - kill connection with open cursor"
        admin_conn = test_env.get_admin_connection()
        conn = test_env.get_connection()
        self.assertEqual(conn.is_healthy(), True)
        sid, serial = self.get_sid_serial(conn)
        with admin_conn.cursor() as admin_cursor:
            sql = f"alter system kill session '{sid},{serial}'"
            admin_cursor.execute(sql)
        with self.assertRaisesFullCode("DPY-4011"):
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
        self.assertFalse(conn.is_healthy())

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_4348(self):
        "4348 - kill connection in cursor context manager"
        admin_conn = test_env.get_admin_connection()
        conn = test_env.get_connection()
        self.assertEqual(conn.is_healthy(), True)
        sid, serial = self.get_sid_serial(conn)
        with admin_conn.cursor() as admin_cursor:
            admin_cursor.execute(f"alter system kill session '{sid},{serial}'")
        with self.assertRaisesFullCode("DPY-4011"):
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
        self.assertEqual(conn.is_healthy(), False)

    def test_4349(self):
        "4349 - fetchmany() with and without parameters"
        sql_part = "select user from dual"
        sql = " union all ".join([sql_part] * 10)
        with self.conn.cursor() as cursor:
            cursor.arraysize = 6
            cursor.execute(sql)
            rows = cursor.fetchmany()
            self.assertEqual(len(rows), cursor.arraysize)
            cursor.execute(sql)
            rows = cursor.fetchmany(size=2)
            self.assertEqual(len(rows), 2)
            cursor.execute(sql)
            rows = cursor.fetchmany(numRows=4)
            self.assertEqual(len(rows), 4)
            cursor.execute(sql)
            with self.assertRaisesFullCode("DPY-2014"):
                cursor.fetchmany(size=2, numRows=4)

    def test_4350(self):
        "4350 - access cursor.rowcount after closing cursor"
        with self.conn.cursor() as cursor:
            cursor.execute("select user from dual")
            cursor.fetchall()
            self.assertEqual(cursor.rowcount, 1)
        self.assertEqual(cursor.rowcount, -1)

    def test_4351(self):
        "4351 - changing bind type with define needed"
        self.cursor.execute("delete from TestClobs")
        row_for_1 = (1, "Short value 1")
        row_for_56 = (56, "Short value 56")
        for data in (row_for_1, row_for_56):
            self.cursor.execute(
                "insert into TestClobs (IntCol, ClobCol) values (:1, :2)",
                data,
            )
        sql = "select IntCol, ClobCol from TestClobs where IntCol = :int_col"
        with test_env.DefaultsContextManager("fetch_lobs", False):
            self.cursor.execute(sql, int_col="1")
            self.assertEqual(self.cursor.fetchone(), row_for_1)
            self.cursor.execute(sql, int_col="56")
            self.assertEqual(self.cursor.fetchone(), row_for_56)
            self.cursor.execute(sql, int_col=1)
            self.assertEqual(self.cursor.fetchone(), row_for_1)

    def test_4352(self):
        "4352 - test calling cursor.parse() twice with the same statement"
        self.cursor.execute("truncate table TestTempTable")
        data = (4363, "Value for test 4363")
        self.cursor.execute(
            "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)",
            data,
        )
        sql = "update TestTempTable set StringCol1 = :v where IntCol = :i"
        for i in range(2):
            self.cursor.parse(sql)
            self.cursor.execute(sql, ("Updated value", data[0]))

    def test_4353(self):
        "4353 - test addition of column to cached query"
        table_name = "test_4365"
        try:
            self.cursor.execute(f"drop table {table_name}")
        except oracledb.DatabaseError:
            pass
        data = ("val 1", "val 2")
        self.cursor.execute(f"create table {table_name} (col1 varchar2(10))")
        self.cursor.execute(f"insert into {table_name} values (:1)", [data[0]])
        self.conn.commit()
        self.cursor.execute(f"select * from {table_name}")
        self.assertEqual(self.cursor.fetchall(), [(data[0],)])
        self.cursor.execute(f"alter table {table_name} add col2 varchar2(10)")
        self.cursor.execute(f"update {table_name} set col2 = :1", [data[1]])
        self.conn.commit()
        self.cursor.execute(f"select * from {table_name}")
        self.assertEqual(self.cursor.fetchall(), [data])

    def test_4354(self):
        "4354 - test population of array var with too many elements"
        var = self.cursor.arrayvar(int, 3)
        with self.assertRaisesFullCode("DPY-2016"):
            var.setvalue(0, [1, 2, 3, 4])

    def test_4355(self):
        "4355 - test executemany() with PL/SQL and increasing data lengths"
        sql = "begin :1 := length(:2); end;"
        var = self.cursor.var(int, arraysize=3)
        self.cursor.executemany(
            sql, [(var, "one"), (var, "two"), (var, "end")]
        )
        self.assertEqual(var.values, [3, 3, 3])
        self.cursor.executemany(
            sql, [(var, "three"), (var, "four"), (var, "end")]
        )
        self.assertEqual(var.values, [5, 4, 3])
        self.cursor.executemany(
            sql, [(var, "five"), (var, "six"), (var, "end")]
        )
        self.assertEqual(var.values, [4, 3, 3])

    def test_4356(self):
        "4356 - test cursor.rowcount values for queries"
        max_rows = 93
        self.cursor.arraysize = 10
        self.cursor.execute(
            "select rownum as id from dual connect by rownum <= :1",
            [max_rows],
        )
        self.assertEqual(self.cursor.rowcount, 0)
        batch_num = 1
        while True:
            rows = self.cursor.fetchmany()
            if not rows:
                break
            expected_value = min(max_rows, batch_num * self.cursor.arraysize)
            self.assertEqual(self.cursor.rowcount, expected_value)
            batch_num += 1
        self.cursor.fetchall()
        self.assertEqual(self.cursor.rowcount, max_rows)

    def test_4357(self):
        "4357 - test bind order for PL/SQL"
        self.cursor.execute("delete from TestClobs")
        sql = """
            insert into TestClobs (IntCol, CLOBCol, ExtraNumCol1)
            values (:1, :2, :3)"""
        data = "x" * 9000
        rows = [(1, data, 5), (2, data, 6)]
        self.cursor.execute(sql, rows[0])
        plsql = f"begin {sql}; end;"
        self.cursor.execute(plsql, rows[1])
        self.conn.commit()
        with test_env.DefaultsContextManager("fetch_lobs", False):
            self.cursor.execute(
                """
                select IntCol, CLOBCol, ExtraNumCol1
                from TestCLOBs
                order by IntCol
                """
            )
            self.assertEqual(self.cursor.fetchall(), rows)

    def test_4358(self):
        "4358 - test rebuild of table with LOB in cached query (as string)"
        table_name = "test_4370"
        drop_sql = f"drop table {table_name} purge"
        create_sql = f"""
            create table {table_name} (
                Col1 number(9) not null,
                Col2 clob not null
            )"""
        insert_sql = f"insert into {table_name} values (:1, :2)"
        query_sql = f"select * from {table_name} order by Col1"
        data = [(1, "CLOB value 1"), (2, "CLOB value 2")]
        try:
            self.cursor.execute(drop_sql)
        except oracledb.DatabaseError:
            pass
        with test_env.DefaultsContextManager("fetch_lobs", False):
            self.cursor.execute(create_sql)
            self.cursor.executemany(insert_sql, data)
            self.cursor.execute(query_sql)
            self.assertEqual(self.cursor.fetchall(), data)
            self.cursor.execute(query_sql)
            self.assertEqual(self.cursor.fetchall(), data)
            self.cursor.execute(drop_sql)
            self.cursor.execute(create_sql)
            self.cursor.executemany(insert_sql, data)
            self.cursor.execute(query_sql)
            self.assertEqual(self.cursor.fetchall(), data)

    def test_4359(self):
        "4359 - test rebuild of table with LOB in cached query (as LOB)"
        table_name = "test_4371"
        drop_sql = f"drop table {table_name} purge"
        create_sql = f"""
            create table {table_name} (
                Col1 number(9) not null,
                Col2 clob not null)"""
        insert_sql = f"insert into {table_name} values (:1, :2)"
        query_sql = f"select * from {table_name} order by Col1"
        data = [(1, "CLOB value 1"), (2, "CLOB value 2")]
        try:
            self.cursor.execute(drop_sql)
        except oracledb.DatabaseError:
            pass
        self.cursor.execute(create_sql)
        self.cursor.executemany(insert_sql, data)
        self.cursor.execute(query_sql)
        fetched_data = [(n, c.read()) for n, c in self.cursor]
        self.assertEqual(fetched_data, data)
        self.cursor.execute(query_sql)
        fetched_data = [(n, c.read()) for n, c in self.cursor]
        self.assertEqual(fetched_data, data)
        self.cursor.execute(drop_sql)
        self.cursor.execute(create_sql)
        self.cursor.executemany(insert_sql, data)
        self.cursor.execute(query_sql)
        fetched_data = [(n, c.read()) for n, c in self.cursor]
        self.assertEqual(fetched_data, data)

    @unittest.skipIf(
        test_env.get_server_version() <= (12, 2), "unsupported database"
    )
    @unittest.skipIf(
        test_env.get_client_version() <= (12, 2), "unsupported database"
    )
    def test_4360(self):
        "4360 - fetch JSON columns as Python objects"
        expected_data = [
            (1, [1, 2, 3], [4, 5, 6], [7, 8, 9]),
            (2, None, None, None),
        ]
        self.cursor.execute("select * from TestJsonCols order by IntCol")
        self.assertEqual(self.cursor.fetchall(), expected_data)

    @unittest.skipIf(
        test_env.get_server_version() < (23, 1), "unsupported database"
    )
    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "unsupported client"
    )
    def test_4361(self):
        "4361 - fetch table with domain and annotations"
        self.cursor.execute("select * from TableWithDomainAndAnnotations")
        self.assertEqual(self.cursor.fetchall(), [(1, 25)])
        column_1 = self.cursor.description[0]
        self.assertIsNone(column_1.domain_schema)
        self.assertIsNone(column_1.domain_name)
        self.assertIsNone(column_1.annotations)
        column_2 = self.cursor.description[1]
        self.assertEqual(
            column_2.domain_schema, test_env.get_main_user().upper()
        )
        self.assertEqual(column_2.domain_name, "SIMPLEDOMAIN")
        expected_annotations = {
            "ANNO_1": "first annotation",
            "ANNO_2": "second annotation",
            "ANNO_3": "",
        }
        self.assertEqual(column_2.annotations, expected_annotations)

    def test_4362(self):
        "4362 - test getting statement after it was executed"
        cursor = self.conn.cursor()
        sql = "select 1 from dual"
        cursor.execute(sql)
        self.assertEqual(cursor.statement, sql)

    def test_4363(self):
        "4363 - test getting cursor fetchvars"
        cursor = self.conn.cursor()
        self.assertIsNone(cursor.fetchvars)

        cursor.execute("truncate table TestTempTable")
        cursor.execute(
            "insert into TestTempTable (IntCol, StringCol1) values (1, '12')",
        )
        cursor.execute("select IntCol, StringCol1 from TestTempTable")
        cursor.fetchall()
        self.assertEqual(len(cursor.fetchvars), 2)
        self.assertEqual(cursor.fetchvars[0].getvalue(), 1)
        self.assertEqual(cursor.fetchvars[1].getvalue(), "12")

    def test_4364(self):
        "4364 - test fetchmany() with non-default cursor.arraysize"
        self.cursor.arraysize = 20
        values = [(i,) for i in range(30)]
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.executemany(
            "insert into TestTempTable (IntCol) values (:1)", values
        )
        self.cursor.execute("select IntCol from TestTempTable order by IntCol")
        # fetch first 20 elements
        fetched_values = self.cursor.fetchmany()
        self.assertEqual(fetched_values, values[: self.cursor.arraysize])

        # fetch missing elements
        fetched_values = self.cursor.fetchmany()
        self.assertEqual(fetched_values, values[self.cursor.arraysize :])

    def test_4365(self):
        "4365 - negative tests for cursor.arraysize"
        with self.assertRaisesFullCode("DPY-2045"):
            self.cursor.arraysize = 0
        with self.assertRaisesFullCode("DPY-2045"):
            self.cursor.arraysize = -1
        with self.assertRaisesFullCode("DPY-2045"):
            self.cursor.arraysize = "not valid"

    def test_4366(self):
        "4366 - test fetching LOBs after an error"
        sql = """
            select
                to_clob(:val),
                1 / (dbms_lob.getlength(to_clob(:val)) - 1)
            from dual"""
        with self.assertRaisesFullCode("ORA-01476"):
            self.cursor.execute(sql, val="a")
            self.cursor.fetchall()
        self.cursor.execute(sql, val="bb")
        lob, num_val = self.cursor.fetchone()
        self.assertEqual(lob.read(), "bb")
        self.assertEqual(num_val, 1)

    def test_4367(self):
        "4367 - test parse() with autocommit enabled"
        conn = test_env.get_connection()
        conn.autocommit = True
        cursor = conn.cursor()
        cursor.execute("truncate table TestTempTable")
        cursor.parse("insert into TestTempTable (IntCol) values (:1)")
        cursor.execute(None, [1])

    def test_4368(self):
        "4368 - test cursor.setinputsizes() with early failed execute"
        self.cursor.setinputsizes(a=int, b=str)
        with self.assertRaisesFullCode("DPY-2006"):
            self.cursor.execute("select :c from dual", [5])
        value = 4368
        self.cursor.execute("select :d from dual", [value])
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(fetched_value, value)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
7200 - Module for testing parsing of tnsnames.ora files.
"""

import os
import tempfile

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    requires_connection = False

    def test_7200(self):
        "7200 - test simple tnsnames.ora entry"
        host = "host_7200"
        port = 7200
        service_name = "service_7200"
        network_service_name = "nsn_7200"
        connect_string = f"""
            (DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})(PORT={port}))
            (CONNECT_DATA=(SERVICE_NAME={service_name})))"""
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string}")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
        self.assertEqual(params.host, host)
        self.assertEqual(params.port, port)
        self.assertEqual(params.service_name, service_name)

    def test_7201(self):
        "7201 - test missing entry in tnsnames.ora"
        with tempfile.TemporaryDirectory() as temp_dir:
            params = oracledb.ConnectParams(config_dir=temp_dir)
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write("# no entries")
            with self.assertRaisesFullCode("DPY-4000"):
                params.parse_connect_string("nsn_7201")
            self.assertEqual(params.get_network_service_names(), [])

    def test_7202(self):
        "7202 - test missing tnsnames.ora in configuration directory"
        with tempfile.TemporaryDirectory() as temp_dir:
            params = oracledb.ConnectParams(config_dir=temp_dir)
            with self.assertRaisesFullCode("DPY-4026"):
                params.parse_connect_string("nsn_7202")
            with self.assertRaisesFullCode("DPY-4026"):
                params.get_network_service_names()

    def test_7203(self):
        "7203 - test tnsnames.ora with invalid entries"
        host = "host_7203"
        port = 7203
        service_name = "service_7203"
        connect_string = f"""
            (DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})(PORT={port}))
            (CONNECT_DATA=(SERVICE_NAME={service_name})))"""
        network_service_name = "nsn_7203"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write("some garbage data which is not a valid entry\n")
                f.write(f"{network_service_name} = {connect_string}\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
            self.assertEqual(params.host, host)
            self.assertEqual(params.port, port)
            self.assertEqual(params.service_name, service_name)

    def test_7204(self):
        "7204 - test tnsnames.ora with multiple aliases on one line"
        host = "host_7204"
        port = 7204
        service_name = "service_7204"
        connect_string = f"""
            (DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})(PORT={port}))
            (CONNECT_DATA=(SERVICE_NAME={service_name})))"""
        network_service_names = "nsn_7204a,nsn_7204b"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(f"{network_service_names} = {connect_string}\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            for name in network_service_names.split(","):
                params.parse_connect_string(name)
                self.assertEqual(params.host, host)
                self.assertEqual(params.port, port)
                self.assertEqual(params.service_name, service_name)
            self.assertEqual(
                params.get_network_service_names(),
                network_service_names.upper().split(","),
            )

    def test_7205(self):
        "7205 - test easy connect string in tnsnames.ora"
        host = "host_7205"
        port = 7205
        service_name = "service_7205"
        connect_string = f"tcp://{host}:{port}/{service_name}"
        network_service_name = "nsn_7205"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string}")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
            self.assertEqual(
                params.get_network_service_names(),
                [network_service_name.upper()],
            )
        self.assertEqual(params.host, host)
        self.assertEqual(params.port, port)
        self.assertEqual(params.service_name, service_name)

    def test_7206(self):
        "7206 - parse connect descriptor with / character in tnsnames.ora"
        host = "host_7206"
        port = 7206
        service_name = "service_7206"
        wallet_location = "/some/dir/7206"
        connect_string = f"""
            (DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})(PORT={port}))
            (CONNECT_DATA=(SERVICE_NAME={service_name}))
            (SECURITY=(MY_WALLET_DIRECTORY={wallet_location})))"""
        network_service_name = "nsn_7206"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string}")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
        self.assertEqual(params.host, host)
        self.assertEqual(params.port, port)
        self.assertEqual(params.service_name, service_name)
        self.assertEqual(params.wallet_location, wallet_location)

    def test_7207(self):
        "7207 - parse IFILE with files in same directory"
        host_a = "host_7207a"
        host_b = "host_7207b"
        port_a = 72071
        port_b = 72072
        service_name_a = "service_7207a"
        service_name_b = "service_7207b"
        connect_string_a = f"{host_a}:{port_a}/{service_name_a}"
        connect_string_b = f"{host_b}:{port_b}/{service_name_b}"
        network_service_name_a = "nsn_7207a"
        network_service_name_b = "nsn_7207b"
        include_file_name = "inc_7207.ora"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, include_file_name)
            with open(file_name, "w") as f:
                f.write(f"{network_service_name_b} = {connect_string_b}")
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(f"{network_service_name_a} = {connect_string_a}\n")
                f.write(f"ifile = {include_file_name}")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name_a)
            self.assertEqual(params.host, host_a)
            self.assertEqual(params.port, port_a)
            self.assertEqual(params.service_name, service_name_a)
            params.parse_connect_string(network_service_name_b)
            self.assertEqual(params.host, host_b)
            self.assertEqual(params.port, port_b)
            self.assertEqual(params.service_name, service_name_b)
            self.assertEqual(
                params.get_network_service_names(),
                [
                    network_service_name_a.upper(),
                    network_service_name_b.upper(),
                ],
            )

    def test_7208(self):
        "7208 - parse IFILE with files in different directories"
        host_a = "host_7208a"
        host_b = "host_7208b"
        port_a = 72081
        port_b = 72082
        service_name_a = "service_7208a"
        service_name_b = "service_7208b"
        connect_string_a = f"{host_a}:{port_a}/{service_name_a}"
        connect_string_b = f"{host_b}:{port_b}/{service_name_b}"
        network_service_name_a = "nsn_7208a"
        network_service_name_b = "nsn_7208b"
        include_file_name = "inc_7208.ora"
        dir_1 = tempfile.TemporaryDirectory()
        dir_2 = tempfile.TemporaryDirectory()
        with dir_1 as primary_temp_dir, dir_2 as included_temp_dir:
            file_name = os.path.join(included_temp_dir, include_file_name)
            with open(file_name, "w") as f:
                f.write(f"{network_service_name_b} = {connect_string_b}")
            primary_file_name = os.path.join(primary_temp_dir, "tnsnames.ora")
            with open(primary_file_name, "w") as f:
                f.write(f"{network_service_name_a} = {connect_string_a}\n")
                f.write(f"ifile = {file_name}")
            params = oracledb.ConnectParams(config_dir=primary_temp_dir)
            params.parse_connect_string(network_service_name_a)
            self.assertEqual(params.host, host_a)
            self.assertEqual(params.port, port_a)
            self.assertEqual(params.service_name, service_name_a)
            params.parse_connect_string(network_service_name_b)
            self.assertEqual(params.host, host_b)
            self.assertEqual(params.port, port_b)
            self.assertEqual(params.service_name, service_name_b)
            self.assertEqual(
                params.get_network_service_names(),
                [
                    network_service_name_a.upper(),
                    network_service_name_b.upper(),
                ],
            )

    def test_7209(self):
        "7209 - cycle detection in same file"
        with tempfile.TemporaryDirectory() as temp_dir:
            network_service_name = "nsn_7209"
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(f"{network_service_name} = some_host/some_service\n")
                f.write("IFILE = tnsnames.ora")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            with self.assertRaisesFullCode("DPY-4030"):
                params.parse_connect_string(network_service_name)

    def test_7210(self):
        "7210 - cycle detection in directly included file"
        with tempfile.TemporaryDirectory() as temp_dir:
            network_service_name = "nsn_7210"
            include_name = "included_7210.ora"
            primary_file_name = os.path.join(temp_dir, "tnsnames.ora")
            include_file_name = os.path.join(temp_dir, include_name)
            with open(primary_file_name, "w") as f:
                f.write(f"{network_service_name} = some_host/some_service\n")
                f.write(f"IFILE = {include_name}")
            with open(include_file_name, "w") as f:
                f.write("IFILE = tnsnames.ora")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            with self.assertRaisesFullCode("DPY-4030"):
                params.parse_connect_string(network_service_name)

    def test_7211(self):
        "7211 - cycle detection in indirectly included file"
        with tempfile.TemporaryDirectory() as temp_dir:
            network_service_name = "nsn_7211"
            include_name_a = "included_7211_a.ora"
            include_name_b = "included_7211_b.ora"
            primary_file_name = os.path.join(temp_dir, "tnsnames.ora")
            include_file_name_a = os.path.join(temp_dir, include_name_a)
            include_file_name_b = os.path.join(temp_dir, include_name_b)
            with open(primary_file_name, "w") as f:
                f.write(f"{network_service_name} = some_host/some_service\n")
                f.write(f"IFILE = {include_name_a}")
            with open(include_file_name_a, "w") as f:
                f.write(f"IFILE = {include_name_b}")
            with open(include_file_name_b, "w") as f:
                f.write("IFILE = tnsnames.ora")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            with self.assertRaisesFullCode("DPY-4030"):
                params.parse_connect_string(network_service_name)

    def test_7212(self):
        "7212 - duplicate entry in same file, but identical connect strings"
        host = "host_7212"
        port = 7212
        service_name = "service_7212"
        connect_string = f"{host}:{port}/{service_name}"
        network_service_name = "nsn_7212"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string}\n")
                f.write("some_other_nsn = some_host/some_service\n")
                f.write(f"{network_service_name} = {connect_string}\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
            self.assertEqual(params.host, host)
            self.assertEqual(params.port, port)
            self.assertEqual(params.service_name, service_name)

    def test_7213(self):
        "7213 - duplicate entry in same file, but different connect strings"
        host_a = "host_7213a"
        port = 7213
        service_name_a = "service_7213a"
        host_b = "host_7213b"
        service_name_b = "service_7213b"
        connect_string_a = f"{host_a}:{port}/{service_name_a}"
        connect_string_b = f"{host_b}:{port}/{service_name_b}"
        network_service_name = "nsn_7213"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string_a}\n")
                f.write("some_other_nsn = some_host/some_service\n")
                f.write(f"{network_service_name} = {connect_string_b}\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
            self.assertEqual(params.host, host_b)
            self.assertEqual(params.port, port)
            self.assertEqual(params.service_name, service_name_b)

    def test_7214(self):
        "7214 - duplicate entry in other file, but identical connect strings"
        host = "host_7214"
        port = 7214
        service_name = "service_7214"
        connect_string = f"{host}:{port}/{service_name}"
        network_service_name = "nsn_7214"
        include_name = "inc_7214.ora"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            include_file_name = os.path.join(temp_dir, include_name)
            with open(file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string}\n")
                f.write(f"IFILE = {include_name}")
            with open(include_file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string}\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
            self.assertEqual(params.host, host)
            self.assertEqual(params.port, port)
            self.assertEqual(params.service_name, service_name)

    def test_7215(self):
        "7215 - duplicate entry in other file, but different connect strings"
        host_a = "host_7215a"
        port = 7215
        service_name_a = "service_7215a"
        host_b = "host_7215b"
        service_name_b = "service_7215b"
        connect_string_a = f"{host_a}:{port}/{service_name_a}"
        connect_string_b = f"{host_b}:{port}/{service_name_b}"
        network_service_name = "nsn_7215"
        include_name = "inc_7215.ora"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            include_file_name = os.path.join(temp_dir, include_name)
            with open(file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string_a}\n")
                f.write(f"IFILE = {include_name}")
            with open(include_file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string_b}\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
            self.assertEqual(params.host, host_b)
            self.assertEqual(params.port, port)
            self.assertEqual(params.service_name, service_name_b)

    def test_7216(self):
        "7216 - test missing IFILE in tnsnames.ora"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write("IFILE = missing.ora\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            with self.assertRaisesFullCode("DPY-4026"):
                params.parse_connect_string("anything")

    def test_7217(self):
        "7217 - test duplicate IFILE, same file"
        host = "host_7217"
        port = 7217
        service_name = "service_7217"
        connect_string = f"{host}:{port}/{service_name}"
        network_service_name = "nsn_7217"
        include_name = "inc_7217.ora"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            include_file_name = os.path.join(temp_dir, include_name)
            with open(file_name, "w") as f:
                f.write(f"IFILE = {include_name}\n")
                f.write("another_nsn = some_host/some_service\n")
                f.write(f"IFILE = {include_name}\n")
            with open(include_file_name, "w") as f:
                f.write(f"{network_service_name} = {connect_string}\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
            self.assertEqual(params.host, host)
            self.assertEqual(params.port, port)
            self.assertEqual(params.service_name, service_name)
            self.assertEqual(
                params.get_network_service_names(),
                [network_service_name.upper(), "ANOTHER_NSN"],
            )

    def test_7218(self):
        "7218 - test duplicate IFILE, different files"
        host = "host_7218"
        port = 7218
        service_name = "service_7218"
        connect_string = f"{host}:{port}/{service_name}"
        network_service_name = "nsn_7218"
        include_name_a = "inc_7218_a.ora"
        include_name_b = "inc_7218_b.ora"
        include_name_c = "inc_7218_c.ora"
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            include_file_name_a = os.path.join(temp_dir, include_name_a)
            include_file_name_b = os.path.join(temp_dir, include_name_b)
            include_file_name_c = os.path.join(temp_dir, include_name_c)
            with open(file_name, "w") as f:
                f.write(f"IFILE = {include_name_a}\n")
                f.write("another_nsn = some_host/some_service\n")
                f.write(f"IFILE = {include_name_b}\n")
            with open(include_file_name_a, "w") as f:
                f.write("in_a = some_host/some_service\n")
                f.write(f"IFILE = {include_name_c}\n")
            with open(include_file_name_b, "w") as f:
                f.write("in_b = some_host/some_service\n")
                f.write(f"IFILE = {include_name_c}\n")
            with open(include_file_name_c, "w") as f:
                f.write(f"{network_service_name} = {connect_string}\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
            self.assertEqual(params.host, host)
            self.assertEqual(params.port, port)
            self.assertEqual(params.service_name, service_name)
            self.assertEqual(
                params.get_network_service_names(),
                ["IN_A", network_service_name.upper(), "ANOTHER_NSN", "IN_B"],
            )

    def test_7219(self):
        "7219 - test tnsnames.ora with multiple aliases on different lines"
        host = "host_7219"
        port = 7219
        service_name = "service_7219"
        connect_string = f"{host}:{port}/{service_name}"
        network_service_names = ["nsn_7219a", "nsn_7219b", "nsn_7219c"]
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(",\n".join(network_service_names))
                f.write(f" = {connect_string}")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            for name in network_service_names:
                params.parse_connect_string(name)
                self.assertEqual(params.host, host)
                self.assertEqual(params.port, port)
                self.assertEqual(params.service_name, service_name)
            self.assertEqual(
                params.get_network_service_names(),
                [n.upper() for n in network_service_names],
            )

    def test_7220(self):
        "7220 - test tnsnames.ora with comment embedded in dsn"
        host = "host_7220"
        port = 7220
        service_name = "service_7220"
        network_service_name = "nsn_7220"
        connect_string = f"""
            (DESCRIPTION=
                (ADDRESS=(PROTOCOL=TCP)(HOST={host})(PORT={port}))
                (CONNECT_DATA=
                    (SERVICE_NAME={service_name})
                    # embedded comment
                )
            )"""
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                print(f"{network_service_name} = {connect_string}", file=f)
            params = oracledb.ConnectParams(config_dir=temp_dir)
            params.parse_connect_string(network_service_name)
            self.assertEqual(params.host, host)
            self.assertEqual(params.port, port)
            self.assertEqual(params.service_name, service_name)
            self.assertEqual(
                params.get_network_service_names(),
                [network_service_name.upper()],
            )

    def test_7221(self):
        "7221 - test tnsnames.ora with a comment between aliases"
        test_values = [
            ("nsn_7221_1", "tcp://host_7221:7221/service_7222_1"),
            ("nsn_7221_2", "tcp://host_7222:7222/service_7222_2"),
        ]
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            for i in range(3):
                entries = [f"{n} = {c}\n" for n, c in test_values]
                entries.insert(i, "# COMMENT \n")
                with open(file_name, "w") as f:
                    f.writelines(entries)
                params = oracledb.ConnectParams(config_dir=temp_dir)
                self.assertEqual(
                    params.get_network_service_names(),
                    [n.upper() for n, _ in test_values],
                )

    def test_7222(self):
        "7222 - test tnsnames.ora with easy connect and connect descriptors"
        network_service_name1 = "nsn_7222_1"
        connect_string1 = """
            (DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST=host_7220)(PORT=7222))
            (CONNECT_DATA=(SERVICE_NAME=service_7222_1)))"""

        network_service_name2 = "nsn_7222_2"
        connect_string2 = "tcp://host_7222:7222/service_7222_2"

        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "tnsnames.ora")
            with open(file_name, "w") as f:
                f.write(f"{network_service_name1} = {connect_string1}\n")
                f.write(f"{network_service_name2} = {connect_string2}\n")
            params = oracledb.ConnectParams(config_dir=temp_dir)
            self.assertEqual(
                params.get_network_service_names(),
                [network_service_name1.upper(), network_service_name2.upper()],
            )


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
7700 - Module for testing the VECTOR database type with storage type SPARSE
available in Oracle Database 23.7 and higher.
"""

import array
import json
import unittest
import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_client_version() >= (23, 7), "unsupported client"
)
@unittest.skipUnless(
    test_env.get_server_version() >= (23, 7), "unsupported server"
)
class TestCase(test_env.BaseTestCase):
    def __test_insert_and_fetch(self, vector, column_name, expected_typecode):
        """
        Test inserting sparse and fetching from a dense vector column.
        """
        self.cursor.execute("delete from TestVectors")
        self.cursor.execute(
            f"""
            insert into TestVectors (IntCol, {column_name})
            values(1, :vector)
            """,
            vector=vector,
        )
        self.conn.commit()
        self.cursor.execute(f"select {column_name} from TestVectors")
        (fetched_value,) = self.cursor.fetchone()
        dense_values = [0 for _ in range(vector.num_dimensions)]
        for i, index in enumerate(vector.indices):
            if expected_typecode == "b":
                dense_values[index] = int(vector.values[i])
            else:
                dense_values[index] = vector.values[i]
        expected_value = array.array(expected_typecode, dense_values)
        self.assertEqual(fetched_value, expected_value)
        self.assertEqual(fetched_value.typecode, expected_typecode)

    def __test_insert_and_fetch_sparse(
        self, vector, column_name, expected_typecode
    ):
        """
        Test inserting and fetching from a sparse vector column.
        """
        self.cursor.execute("delete from TestSparseVectors")
        self.cursor.execute(
            f"""
            insert into TestSparseVectors (IntCol, {column_name})
            values(1, :vector)
            """,
            vector=vector,
        )
        self.conn.commit()
        self.cursor.execute(f"select {column_name} from TestSparseVectors")
        (fetched_value,) = self.cursor.fetchone()
        expected_value = vector.values
        if fetched_value.values.typecode == "b":
            expected_value = array.array("b", [int(i) for i in vector.values])
        expected_indices = vector.indices
        expected_num_dimensions = vector.num_dimensions
        self.assertEqual(fetched_value.values, expected_value)
        self.assertEqual(fetched_value.indices, expected_indices)
        self.assertEqual(fetched_value.num_dimensions, expected_num_dimensions)

    def __fetch_with_vector(
        self,
        vector,
        column_name,
        dimensions,
        vector_format,
        expected_typecode,
    ):
        """
        Test fetching a vector with vector() function.
        """
        self.cursor.execute("delete from TestSparseVectors")
        self.cursor.execute(
            f"""
            insert into TestSparseVectors (IntCol, {column_name})
            values(1, :vector)
            """,
            vector=vector,
        )
        self.cursor.execute(
            f"""
            select
            vector({column_name}, {dimensions}, {vector_format}, DENSE)
            from TestSparseVectors
            """
        )
        (fetched_value,) = self.cursor.fetchone()
        self.assertIsInstance(fetched_value, array.array)
        self.assertEqual(fetched_value.typecode, expected_typecode)

        self.cursor.execute(
            f"""
            select
            vector({column_name}, {dimensions}, {vector_format}, SPARSE)
            from TestSparseVectors
            """
        )
        (fetched_value,) = self.cursor.fetchone()
        self.assertIsInstance(fetched_value, oracledb.SparseVector)
        self.assertEqual(fetched_value.values.typecode, expected_typecode)

    def test_7700(self):
        "7700 - test binding in a sparse vector with oracledb.SparseVector"
        vector = oracledb.SparseVector(3, [1], [9])
        self.cursor.execute("select :1 from dual", [vector])
        (fetched_value,) = self.cursor.fetchone()
        self.assertIsInstance(fetched_value, oracledb.SparseVector)
        self.assertEqual(fetched_value.num_dimensions, vector.num_dimensions)
        self.assertEqual(fetched_value.indices, vector.indices)
        self.assertEqual(fetched_value.values, vector.values)

    def test_7701(self):
        "7701 - test binding in a sparse vector of type float32"
        vector = oracledb.SparseVector(3, [1], array.array("f", [0.5]))
        self.cursor.execute("select :1 from dual", [vector])
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(fetched_value.values, vector.values)
        self.assertEqual(fetched_value.indices, vector.indices)
        self.assertEqual(fetched_value.num_dimensions, vector.num_dimensions)
        self.assertEqual(fetched_value.values.typecode, "f")

    def test_7702(self):
        "7702 - test binding in a sparse vector of type float64"
        vector = oracledb.SparseVector(3, [1], array.array("d", [0.25]))
        self.cursor.execute("select :1 from dual", [vector])
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(fetched_value.values, vector.values)
        self.assertEqual(fetched_value.indices, vector.indices)
        self.assertEqual(fetched_value.num_dimensions, vector.num_dimensions)
        self.assertEqual(fetched_value.values.typecode, "d")
        self.assertIsInstance(fetched_value, oracledb.SparseVector)

    def test_7703(self):
        "7703 - test binding in a sparse vector of type int8"
        vector = oracledb.SparseVector(3, [1], array.array("b", [3]))
        self.cursor.execute("select :1 from dual", [vector])
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(fetched_value.values, vector.values)
        self.assertEqual(fetched_value.indices, vector.indices)
        self.assertEqual(fetched_value.num_dimensions, vector.num_dimensions)
        self.assertEqual(fetched_value.values.typecode, "b")
        self.assertIsInstance(fetched_value, oracledb.SparseVector)

    def test_7704(self):
        "7704 - insert a float32 sparse vector into a float32 column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [1.5, 0.25, 0.5])
        )
        self.__test_insert_and_fetch(value, "Vector32Col", "f")
        self.__test_insert_and_fetch_sparse(value, "SparseVector32Col", "f")

    def test_7705(self):
        "7705 - insert a float32 vector into a float64 column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("d", [1.5, 0.25, 0.5])
        )
        self.__test_insert_and_fetch(value, "Vector64Col", "d")
        self.__test_insert_and_fetch_sparse(value, "SparseVector64Col", "d")

    def test_7706(self):
        "7706 - insert a float32 vector into a flexible format column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [1.5, 0.25, 0.5])
        )
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "f")
        self.__test_insert_and_fetch_sparse(
            value, "SparseVectorFlexAllCol", "f"
        )

    def test_7707(self):
        "7707 - insert a float64 vector into a float64 column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("d", [1.5, 0.25, 0.5])
        )
        self.__test_insert_and_fetch(value, "Vector64Col", "d")
        self.__test_insert_and_fetch_sparse(value, "SparseVector64Col", "d")

    def test_7708(self):
        "7708 - insert float64 vector into a float32 column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [1.5, 0.25, 0.5])
        )
        self.__test_insert_and_fetch(value, "Vector32Col", "f")
        self.__test_insert_and_fetch_sparse(value, "SparseVector32Col", "f")

    def test_7709(self):
        "7709 - insert float64 vector into a flexible type column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("d", [1.5, 0.25, 0.5])
        )
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "d")
        self.__test_insert_and_fetch_sparse(
            value, "SparseVectorFlexAllCol", "d"
        )

    def test_7710(self):
        "7710 - insert a vector with an invalid size"
        self.cursor.execute("delete from TestVectors")
        self.cursor.execute("delete from TestSparseVectors")
        statements = [
            """
            insert into TestVectors (IntCol, Vector64Col)
            values(1, :1)
            """,
            """
            insert into TestSparseVectors (IntCol, SparseVector64Col)
            values(2, :1)
            """,
        ]
        for statement in statements:
            for num_elems in [4, 20]:
                vector = oracledb.SparseVector(
                    num_elems, [2, 3], array.array("f", [6.54, 9.6])
                )
                with self.assertRaisesFullCode("ORA-51803"):
                    self.cursor.execute(statement, [vector])

    def test_7711(self):
        "7711 - verify fetch info for vectors"
        attr_names = [
            "name",
            "type_code",
            "vector_dimensions",
            "vector_format",
            "vector_is_sparse",
        ]
        expected_values = [
            ["INTCOL", oracledb.DB_TYPE_NUMBER, None, None, None],
            [
                "SPARSEVECTORFLEXALLCOL",
                oracledb.DB_TYPE_VECTOR,
                None,
                None,
                True,
            ],
            [
                "SPARSEVECTORFLEXTYPECOL",
                oracledb.DB_TYPE_VECTOR,
                2,
                None,
                True,
            ],
            [
                "SPARSEVECTORFLEX8COL",
                oracledb.DB_TYPE_VECTOR,
                None,
                oracledb.VECTOR_FORMAT_INT8,
                True,
            ],
            [
                "SPARSEVECTORFLEX32COL",
                oracledb.DB_TYPE_VECTOR,
                None,
                oracledb.VECTOR_FORMAT_FLOAT32,
                True,
            ],
            [
                "SPARSEVECTORFLEX64COL",
                oracledb.DB_TYPE_VECTOR,
                None,
                oracledb.VECTOR_FORMAT_FLOAT64,
                True,
            ],
            [
                "SPARSEVECTOR8COL",
                oracledb.DB_TYPE_VECTOR,
                16,
                oracledb.VECTOR_FORMAT_INT8,
                True,
            ],
            [
                "SPARSEVECTOR32COL",
                oracledb.DB_TYPE_VECTOR,
                16,
                oracledb.VECTOR_FORMAT_FLOAT32,
                True,
            ],
            [
                "SPARSEVECTOR64COL",
                oracledb.DB_TYPE_VECTOR,
                16,
                oracledb.VECTOR_FORMAT_FLOAT64,
                True,
            ],
        ]
        self.cursor.execute("select * from TestSparseVectors")
        values = [
            [getattr(i, n) for n in attr_names]
            for i in self.cursor.description
        ]
        self.assertEqual(values, expected_values)
        self.assertIs(
            self.cursor.description[6].vector_format,
            oracledb.VectorFormat.INT8,
        )

    def test_7712(self):
        "7712 - insert an int8 vector into an int8 column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [1, 0, 5])
        )
        self.__test_insert_and_fetch(value, "Vector8Col", "b")
        self.__test_insert_and_fetch_sparse(value, "SparseVector8Col", "b")

    def test_7713(self):
        "7713 - insert an int8 vector into a float32 column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [1, 0, 5])
        )
        self.__test_insert_and_fetch(value, "Vector32Col", "f")
        self.__test_insert_and_fetch_sparse(value, "SparseVector32Col", "f")

    def test_7714(self):
        "7714 - insert an int8 vector into a float64 column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("b", [1, 0, 5])
        )
        self.__test_insert_and_fetch(value, "Vector64Col", "d")
        self.__test_insert_and_fetch_sparse(value, "SparseVector64Col", "d")

    def test_7715(self):
        "7715 - insert an int8 vector into a flexible column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("b", [1, 0, 5])
        )
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "b")
        self.__test_insert_and_fetch_sparse(
            value, "SparseVectorFlexAllCol", "b"
        )

    def test_7716(self):
        "7716 - insert a float32 vector into an int8 column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [1, 0, 5])
        )
        self.__test_insert_and_fetch(value, "Vector8Col", "b")
        self.__test_insert_and_fetch_sparse(value, "SparseVector8Col", "b")

    def test_7717(self):
        "7717 - insert a float64 vector into an int8 column"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("d", [1, 0, 5])
        )
        self.__test_insert_and_fetch(value, "Vector8Col", "b")
        self.__test_insert_and_fetch_sparse(value, "SparseVector8Col", "b")

    def test_7718(self):
        "7718 - test dml returning vector type"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [1, 0, 5])
        )
        out_var = self.cursor.var(oracledb.DB_TYPE_VECTOR)
        self.cursor.execute("delete from TestSparseVectors")
        self.cursor.execute(
            """
            insert into TestSparseVectors (IntCol, SparseVectorFlex32Col)
            values (1, :value)
            returning SparseVectorFlex32Col into :out_value
            """,
            [value, out_var],
        )
        self.conn.commit()
        vector = out_var.getvalue()[0]
        self.assertEqual(vector.values, value.values)
        self.assertEqual(vector.indices, value.indices)
        self.assertEqual(vector.num_dimensions, value.num_dimensions)

    def test_7719(self):
        "7719 - test handling of NULL vector value"
        self.cursor.execute("delete from TestSparseVectors")
        self.cursor.execute(
            "insert into TestSparseVectors (IntCol) values (1)"
        )
        self.conn.commit()
        self.cursor.execute(
            "select SparseVectorFlexTypeCol from TestSparseVectors"
        )
        (fetched_value,) = self.cursor.fetchone()
        self.assertIsNone(fetched_value)

    def test_7720(self):
        "7720 - insert a float32 vector into an int8 column (negative)"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [-130, 400, 5])
        )
        with self.assertRaisesFullCode("ORA-51806"):
            self.__test_insert_and_fetch(value, "Vector8Col", "b")
        with self.assertRaisesFullCode("ORA-51806"):
            self.__test_insert_and_fetch_sparse(value, "SparseVector8Col", "b")

    def test_7721(self):
        "7721 - insert a float32 vector with 65,533 dimensions"
        value = oracledb.SparseVector(
            65533, [1, 3, 5], array.array("f", [1, 0, 5])
        )
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "f")
        self.__test_insert_and_fetch_sparse(
            value, "SparseVectorFlexAllCol", "f"
        )

    def test_7722(self):
        "7722 - insert vectors with different dimensions"
        for dim in [30, 70, 255, 256, 65534, 65535]:
            for typ in ["f", "d", "b"]:
                with self.subTest(dim=dim, typ=typ):
                    element_value = 3 if typ == "b" else 1.5
                    value = oracledb.SparseVector(
                        dim, [1, 3, 5], array.array(typ, [element_value] * 3)
                    )
                    self.__test_insert_and_fetch(
                        value, "VectorFlexAllCol", typ
                    )
                    self.__test_insert_and_fetch_sparse(
                        value, "SparseVectorFlexAllCol", typ
                    )

    def test_7723(self):
        "7723 - insert and fetch VECTOR data using strings"
        values = [16, [1, 3, 5], [1, 0, 5]]
        vector = oracledb.SparseVector(*values)
        self.cursor.execute("delete from TestSparseVectors")
        self.cursor.execute(
            """
            insert into TestSparseVectors (IntCol, SparseVectorFlexAllCol)
            values(1, :value)
            """,
            value=str(vector),
        )

        def type_handler(cursor, metadata):
            if metadata.name == "SPARSEVECTORFLEXALLCOL":
                return cursor.var(
                    oracledb.DB_TYPE_LONG, arraysize=cursor.arraysize
                )

        self.cursor.outputtypehandler = type_handler

        self.cursor.execute(
            "select SparseVectorFlexAllCol from TestSparseVectors"
        )
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(json.loads(fetched_value), values)

    def test_7724(self):
        "7724 - insert vectors with flexible dimensions and conversion"
        for dim in [30, 255, 256, 257, 32768, 65535]:
            for source_type in ["f", "d", "b"]:
                for target_type in ["f", "d", "b"]:
                    with self.subTest(
                        dim=dim,
                        source_type=source_type,
                        target_type=target_type,
                    ):
                        if target_type == "f":
                            target_col = "VectorFlex32Col"
                        elif target_type == "d":
                            target_col = "VectorFlex64Col"
                        else:
                            target_col = "VectorFlex8Col"
                        element_value = 4 if source_type == "b" else 2.25
                        value = oracledb.SparseVector(
                            dim,
                            [1, 3, 7, 9],
                            array.array(source_type, [element_value] * 4),
                        )
                        self.__test_insert_and_fetch(
                            value, target_col, target_type
                        )
                        self.__test_insert_and_fetch_sparse(
                            value, f"Sparse{target_col}", target_type
                        )

    def test_7725(self):
        "7725 - test binding a vector with inf values (negative)"
        value = oracledb.SparseVector(
            16,
            [1, 3, 5],
            array.array("d", [float("inf"), float("-inf"), float("-inf")]),
        )
        with self.assertRaisesFullCode("ORA-51805", "ORA-51831"):
            self.cursor.execute("select :1 from dual", [value])

    def test_7726(self):
        "7726 - test setting a sparse vector to a vector variable"
        value = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [1, 0, 5])
        )
        var = self.cursor.var(oracledb.DB_TYPE_VECTOR)
        var.setvalue(0, value)
        vector = var.getvalue()
        self.assertEqual(vector.values, value.values)
        self.assertEqual(vector.indices, value.indices)
        self.assertEqual(vector.num_dimensions, value.num_dimensions)

    def test_7727(self):
        "7727 - fetch JSON value with an embedded vector"
        self.cursor.execute("delete from TestSparseVectors")
        vector = oracledb.SparseVector(
            16, [1, 3, 5], array.array("d", [1.5, 0.25, 0.5])
        )
        self.cursor.execute(
            """
            insert into TestSparseVectors (IntCol, SparseVector64Col)
            values (1, :1)
            """,
            [vector],
        )
        self.cursor.execute(
            """
            select json_object(
                'id': 7732,
                'vector' : vector(SparseVector64Col, 16, float64, sparse)
                returning json
            ) from TestSparseVectors
            """
        )
        (result,) = self.cursor.fetchone()
        fetched_vector = result["vector"]
        self.assertIsInstance(fetched_vector, oracledb.SparseVector)
        self.assertEqual(fetched_vector.indices, vector.indices)
        self.assertEqual(fetched_vector.values, vector.values)
        self.assertEqual(fetched_vector.num_dimensions, vector.num_dimensions)

    def test_7728(self):
        "7728 - executemany() without setinputsizes()"
        self.cursor.execute("delete from TestSparseVectors")
        vector = oracledb.SparseVector(
            16, [1, 3, 5], array.array("f", [1, 0, 5])
        )
        values = [vector, [0, 0, 0, 2, 0, 0, 0, 0, 0, 1, 0, 4, 0, 0, 0, 0]]
        self.cursor.executemany(
            """
            insert into TestSparseVectors (IntCol, SparseVector32Col)
            values (:1, :2)
            """,
            list(enumerate(values)),
        )
        self.cursor.execute(
            "select SparseVector32Col from TestSparseVectors order by IntCol"
        )
        (fetched_vector1,), (fetched_vector2,) = self.cursor.fetchall()
        self.assertEqual(fetched_vector1.values, vector.values)
        self.assertEqual(fetched_vector1.indices, vector.indices)
        self.assertEqual(fetched_vector1.num_dimensions, vector.num_dimensions)
        self.assertEqual(
            fetched_vector2.values, array.array("f", [2.0, 1.0, 4.0])
        )
        self.assertEqual(fetched_vector2.indices, array.array("I", [3, 9, 11]))
        self.assertEqual(fetched_vector2.num_dimensions, 16)

    def test_7729(self):
        "7729 - executemany() with setinputsizes()"
        self.cursor.execute("delete from TestSparseVectors")
        vector = oracledb.SparseVector(
            16, [1, 3, 5], array.array("d", [1, 0, 5])
        )
        values = [[144, 0, 1000], vector]
        self.cursor.setinputsizes(None, oracledb.DB_TYPE_VECTOR)
        self.cursor.executemany(
            """
            insert into TestSparseVectors (IntCol, SparseVectorFlex64Col)
            values (:1, :2)
            """,
            list(enumerate(values)),
        )
        self.cursor.execute(
            """
            select SparseVectorFlex64Col
            from TestSparseVectors order by IntCol
            """
        )
        (fetched_vector1,), (fetched_vector2,) = self.cursor.fetchall()
        self.assertEqual(
            fetched_vector1.values, array.array("d", [144.0, 1000.0])
        )
        self.assertEqual(fetched_vector1.indices, array.array("I", [0, 2]))
        self.assertEqual(fetched_vector1.num_dimensions, 3)
        self.assertEqual(fetched_vector2.values, vector.values)
        self.assertEqual(fetched_vector2.indices, vector.indices)
        self.assertEqual(fetched_vector2.num_dimensions, vector.num_dimensions)

    def test_7730(self):
        "7730 - vector with zero dimensions"
        self.cursor.setinputsizes(oracledb.DB_TYPE_VECTOR)
        vector = oracledb.SparseVector(4, [], [])
        with self.assertRaisesFullCode("ORA-51803", "ORA-21560"):
            self.cursor.execute("select :1", [vector])

    def test_7731(self):
        "7731 - test inserting a vector as a string and fetching it"
        self.cursor.execute("delete from TestSparseVectors")
        self.cursor.execute(
            """
            insert into TestSparseVectors (IntCol, SparseVectorFlexAllCol)
            values (1, '[4, [1, 3], [1.0, 2.0]]')
            """
        )
        self.cursor.execute(
            "select SparseVectorFlexAllCol from TestSparseVectors"
        )
        vector = self.cursor.fetchone()[0]
        self.assertEqual(vector.values, array.array("f", [1, 2]))
        self.assertEqual(vector.num_dimensions, 4)
        self.assertEqual(vector.indices, array.array("I", [1, 3]))

    def test_7732(self):
        "7732 - SparseVector() with invalid values"
        # pass strings instead of number or list/array.array
        with self.assertRaises(TypeError):
            oracledb.SparseVector("10", [1, 2], [1.5, 3.5])
        with self.assertRaises(TypeError):
            oracledb.SparseVector(10, "[1, 2]", [1.5, 3.5])
        with self.assertRaises(TypeError):
            oracledb.SparseVector(10, [1, 2], "[1.5, 3.5]")

        # insert matrix
        with self.assertRaises(TypeError):
            oracledb.SparseVector(10, [[1, 2]], [1.5, 3.5])
        with self.assertRaises(TypeError):
            oracledb.SparseVector(10, [1, 2], [[1.5, 3.5]])
        # use num_dimensions as a list
        with self.assertRaises(TypeError):
            oracledb.SparseVector([10], [1, 2], [1.5, 3.5])
        # use num_dimensions as a float
        value = oracledb.SparseVector(10.4, [1, 2], [1.5, 3.5])
        self.assertEqual(value.num_dimensions, 10)

        # negative index
        with self.assertRaises(OverflowError):
            oracledb.SparseVector(10, [-1], [1.5])
        # negative num_dimensions
        with self.assertRaises(OverflowError):
            oracledb.SparseVector(-10, [1], [3.5])
        # use float index
        with self.assertRaises(TypeError):
            oracledb.SparseVector(10, [2.4], [3.5])

    def test_7733(self):
        "7733 - SparseVector() with indices and values of different length"
        with self.assertRaises(TypeError):
            oracledb.SparseVector(10, [1], [1.5, 3.5])
        with self.assertRaises(TypeError):
            oracledb.SparseVector(10, [1, 2, 3, 4], [6.75])

    def test_7734(self):
        "7734 - declare and insert an empty SparseVector"
        value = oracledb.SparseVector(0, [], [])
        self.assertEqual(value.values, array.array("d"))
        self.assertEqual(value.indices, array.array("I"))
        self.assertEqual(value.num_dimensions, 0)
        with self.assertRaisesFullCode("ORA-51803", "ORA-21560", "ORA-51862"):
            self.__test_insert_and_fetch(value, "VectorFlexAllCol", "d")
        with self.assertRaisesFullCode("ORA-51803", "ORA-21560", "ORA-51862"):
            self.__test_insert_and_fetch_sparse(
                value, "SparseVectorFlexAllCol", "d"
            )

    def test_7735(self):
        "7735 - select with vector()"
        dense_vector = array.array(
            "f", [1, 2, 3, 4, 0, 5, 6, 7, 8, 9, 10, 11, 12, 13, 0, 0]
        )
        sparse_vector = oracledb.SparseVector(16, [1], array.array("f", [9]))
        column_names = [
            "SparseVector8Col",
            "SparseVector32Col",
            "SparseVector64Col",
            "SparseVectorFlex8Col",
            "SparseVectorFlex32Col",
            "SparseVectorFlex64Col",
            "SparseVectorFlexAllCol",
        ]
        for vector in [dense_vector, sparse_vector]:
            for column_name in column_names:
                with self.subTest(vector=vector, column_name=column_name):
                    self.__fetch_with_vector(
                        vector, column_name, 16, "INT8", "b"
                    )
                    self.__fetch_with_vector(
                        vector, column_name, 16, "FLOAT32", "f"
                    )
                    self.__fetch_with_vector(
                        vector, column_name, 16, "FLOAT64", "d"
                    )

        # fixed dimension columns
        dense_vector = array.array("f", [1, 2])
        sparse_vector = oracledb.SparseVector(2, [1], array.array("f", [1]))
        for vector in [dense_vector, sparse_vector]:
            for column_name in column_names[3:]:
                with self.subTest(vector=vector, column_name=column_name):
                    self.__fetch_with_vector(
                        vector, column_name, 2, "INT8", "b"
                    )
                    self.__fetch_with_vector(
                        vector, column_name, 2, "FLOAT32", "f"
                    )
                    self.__fetch_with_vector(
                        vector, column_name, 2, "FLOAT64", "d"
                    )

    def test_7736(self):
        "7736 - test from_vector() with returning and vector storage format"
        self.cursor.execute("delete from TestSparseVectors")
        values = [16, [1, 2, 15], [2, 45.5, 73.25]]
        vector = oracledb.SparseVector(*values)
        dense_vector = [0, 2, 45.5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 73.25]
        column_name = "SparseVector64Col"
        self.cursor.execute(
            f"""
            insert into TestSparseVectors (IntCol, {column_name})
            values(1, :vector)
            """,
            vector=vector,
        )
        self.cursor.execute(
            f"""
            select from_vector({column_name} returning clob format sparse)
            from TestSparseVectors
            """
        )
        (lob,) = self.cursor.fetchone()
        self.assertEqual(json.loads(lob.read()), values)
        self.cursor.execute(
            f"""
            select from_vector({column_name} returning clob format dense)
            from TestSparseVectors
            """
        )
        (lob,) = self.cursor.fetchone()
        self.assertEqual(json.loads(lob.read()), dense_vector)
        self.cursor.execute(
            f"""
            select from_vector({column_name} returning clob)
            from TestSparseVectors
            """
        )
        (lob,) = self.cursor.fetchone()
        self.assertEqual(json.loads(lob.read()), values)
        self.cursor.execute(
            f"""
            select from_vector({column_name} returning varchar2 format sparse)
            from TestSparseVectors
            """
        )
        self.assertEqual(json.loads(self.cursor.fetchone()[0]), values)
        self.cursor.execute(
            f"""
            select from_vector({column_name} returning varchar2 format dense)
            from TestSparseVectors
            """
        )
        self.assertEqual(json.loads(self.cursor.fetchone()[0]), dense_vector)
        self.cursor.execute(
            f"""
            select from_vector({column_name} returning varchar2)
            from TestSparseVectors
            """
        )
        self.assertEqual(json.loads(self.cursor.fetchone()[0]), values)


if __name__ == "__main__":
    test_env.run_test_cases()
